@@ -25,15 +25,18 @@ pub use self::value::Value;
 
 pub use self::sql_operator::SQLOperator;
 
+// This could be enhanced to remember the way the identifier was quoted
+pub type SQLIdent = String;
+
 /// SQL Abstract Syntax Tree (AST)
 #[derive(Debug, Clone, PartialEq)]
 pub enum ASTNode {
     /// Identifier e.g. table name or column name
-    SQLIdentifier(String),
+    SQLIdentifier(SQLIdent),
     /// Wildcard e.g. `*`
     SQLWildcard,
     /// Multi part identifier e.g. `myschema.dbo.mytable`
-    SQLCompoundIdentifier(Vec<String>),
+    SQLCompoundIdentifier(Vec<SQLIdent>),
     /// Assigment e.g. `name = 'Fred'` in an UPDATE statement
     SQLAssignment(SQLAssignment),
     /// `IS NULL` expression
@@ -69,12 +72,17 @@ pub enum ASTNode {
         results: Vec<ASTNode>,
         else_result: Option<Box<ASTNode>>,
     },
+    /// A table name or a parenthesized subquery with an optional alias
+    TableFactor {
+        relation: Box<ASTNode>, // SQLNested or SQLCompoundIdentifier
+        alias: Option<SQLIdent>,
+    },
     /// SELECT
     SQLSelect {
         /// projection expressions
         projection: Vec<ASTNode>,
         /// FROM
-        relation: Option<Box<ASTNode>>,
+        relation: Option<Box<ASTNode>>, // TableFactor
         // JOIN
         joins: Vec<Join>,
         /// WHERE
@@ -93,7 +101,7 @@ pub enum ASTNode {
         /// TABLE
         table_name: String,
         /// COLUMNS
-        columns: Vec<String>,
+        columns: Vec<SQLIdent>,
         /// VALUES (vector of rows to insert)
         values: Vec<Vec<ASTNode>>,
     },
@@ -101,7 +109,7 @@ pub enum ASTNode {
         /// TABLE
         table_name: String,
         /// COLUMNS
-        columns: Vec<String>,
+        columns: Vec<SQLIdent>,
         /// VALUES a vector of values to be copied
         values: Vec<Option<String>>,
     },
@@ -188,6 +196,13 @@ impl ToString for ASTNode {
                 }
                 s + " END"
             }
+            ASTNode::TableFactor { relation, alias } => {
+                if let Some(alias) = alias {
+                    format!("{} AS {}", relation.to_string(), alias)
+                } else {
+                    relation.to_string()
+                }
+            }
             ASTNode::SQLSelect {
                 projection,
                 relation,
@@ -366,21 +381,21 @@ impl ToString for SQLAssignment {
 #[derive(Debug, Clone, PartialEq)]
 pub struct SQLOrderByExpr {
     pub expr: Box<ASTNode>,
-    pub asc: bool,
+    pub asc: Option<bool>,
 }
 
 impl SQLOrderByExpr {
-    pub fn new(expr: Box<ASTNode>, asc: bool) -> Self {
+    pub fn new(expr: Box<ASTNode>, asc: Option<bool>) -> Self {
         SQLOrderByExpr { expr, asc }
     }
 }
 
 impl ToString for SQLOrderByExpr {
     fn to_string(&self) -> String {
-        if self.asc {
-            format!("{} ASC", self.expr.as_ref().to_string())
-        } else {
-            format!("{} DESC", self.expr.as_ref().to_string())
+        match self.asc {
+            Some(true) => format!("{} ASC", self.expr.to_string()),
+            Some(false) => format!("{} DESC", self.expr.to_string()),
+            None => self.expr.to_string(),
         }
     }
 }
@@ -388,7 +403,7 @@ impl ToString for SQLOrderByExpr {
 /// SQL column definition
 #[derive(Debug, Clone, PartialEq)]
 pub struct SQLColumnDef {
-    pub name: String,
+    pub name: SQLIdent,
     pub data_type: SQLType,
     pub is_primary: bool,
     pub is_unique: bool,
@@ -417,7 +432,7 @@ impl ToString for SQLColumnDef {
 
 #[derive(Debug, Clone, PartialEq)]
 pub struct Join {
-    pub relation: ASTNode,
+    pub relation: ASTNode, // TableFactor
     pub join_operator: JoinOperator,
 }
 
